import sys
from galprime.isophote_l import Ellipse, EllipseGeometry
from numpy import max, pi, log, unravel_index, argmax, ceil
from photutils.morphology import data_properties


def isophote_fitting(data, config=None, centre_method='standard'):
    """ Wrapper for the photutils isophote extraction routine. Designed to get the most out of profile 
        extractions.

    :param data: Input data (2D cutout)
    :type data: numpy.ndarray
    :param config: The configuration file with isophote fitting params. Will use default parameters if 
        nothing is supplied, defaults to None
    :type config: dict, optional
    :param centre_method: Which method to use when finding the centre for initial isophote, defaults to 'standard'
        'standard'  Uses the central region of the cutout
        'max'       Uses the location of the brightest pixel. Only use if your cutout is properly masked.
    :type centre_method: str, optional
    :return: The photutils isophotelist object containing extraction results.
    :rtype: photutils.isophote.IsophoteList
    """

    fail_count, max_fails = 0, 1000
    linear = False if config is None else config["LINEAR"]
    step = 1. if config is None else config["LINEAR_STEP"]

    # Get centre of image and cutout halfwidth
    if centre_method == 'standard':
        centre = (data.shape[0]/2, data.shape[1]/2)
    elif centre_method == 'max':
        centre = unravel_index(argmax(data), data.shape)
    else:
        centre = (data.shape[0] / 2, data.shape[1] / 2)

    # Here we can just extract out to the requested size if specified in the cutout
    cutout_halfwidth = max((ceil(data.shape[0] / 2), ceil(data.shape[1] / 2)))
    if config is not None:
        cutout_halfwidth = config["EXTRACTION_SIZE"] / 2
    
    fitting_list = []

    # First, try obtaining morphological properties from the data and fit using that starting ellipse
    try:
        morph_cat = data_properties(log(data))
        r = 2.0
        pos = (morph_cat.xcentroid, morph_cat.ycentroid)
        a = morph_cat.semimajor_sigma.value * r
        b = morph_cat.semiminor_sigma.value * r
        theta = morph_cat.orientation.value

        geometry = EllipseGeometry(pos[0], pos[1], sma=a, eps=(1 - (b / a)), pa=theta)
        flux = Ellipse(data, geometry)
        fitting_list = flux.fit_image(maxit=100, maxsma=cutout_halfwidth, step=step, linear=linear,
                                      maxrit=cutout_halfwidth / 2)
        if len(fitting_list) > 0:
            return fitting_list

    except KeyboardInterrupt:
        sys.exit(1)
    except (RuntimeError, ValueError, OverflowError, IndexError):
        fail_count += 1
        if fail_count >= max_fails:
            return []

    # If that fails, test a parameter space of starting ellipses
    try:
        for angle in range(0, 180, 45):
            for sma in range(2, 26, 5):
                for eps in (0.3, 0.5, 0.9):
                    geometry = EllipseGeometry(float(centre[0]), float(centre[1]), eps=eps,
                                               sma=sma, pa=angle * pi / 180.)
                    flux = Ellipse(data, geometry)
                    fitting_list = flux.fit_image(maxsma=cutout_halfwidth, step=step, linear=linear,
                                                  maxrit=cutout_halfwidth / 3)
                    if len(fitting_list) > 0:
                        return fitting_list

    except KeyboardInterrupt:
        sys.exit(1)
    except (RuntimeError, ValueError, OverflowError, IndexError):

        # print("RuntimeError or ValueError")
        fail_count += 1
        if fail_count >= max_fails:
            return []
    except IndexError:
        fail_count += 1
        if fail_count >= max_fails:
            return []

    return fitting_list



import os
import time
import galprime
from astropy.io import fits
from astropy.visualization import ZScaleInterval

from matplotlib import pyplot as plt

import numpy as np
from scipy.signal import convolve2d


class GalPrimeContainer:
    """ A container object for a single "galprime" realization
    """
    def __init__(self, config=None, kde=None, mag_kde=None, background=None, psf=None,
                 model_cutout=None, background_cutout=None, bg_model=None,
                 bgadded_masked=None, bgsub_cutout=None, bgsub_masked=None, 
                 model_profile=None, bgadded_profile=None, bgsub_profile=None,
                 metadata=None):

      
        if config == None:
            self.config = galprime.default_config_params()
        else:
            self.config = config
        
        if metadata is None:
            self.metadata = {}
        else:
            self.metadata = metadata
            
        # Generate Container ID
        self.metadata["ID"] = np.random.randint(1e15,1e16)
        
        self.kde = kde
        self.mag_kde = mag_kde
        self.background = background
        
        # Models without masking
        self.model = model_cutout
        self.convolved_model = None
        self.background_cutout = background_cutout
        self.bgsub_cutout = bgsub_cutout
        self.psf = psf
        
        # Background-added cutouts after masking
        self.bgadded_masked = bgadded_masked
        self.bgsub_masked = bgsub_masked
        self.bg_model = bg_model
        
        # Extracted profiles
        self.model_profile = model_profile
        self.bgadded_profile = bgadded_profile
        self.bgsub_profile = bgsub_profile
    
    # Functions to run ########################################################################
    def generate_model(self):
        if self.config["DEBUG"]:
            print("Generating Sersic model cutout of size", self.config["EXTRACTION_SIZE"])    
        self.model, model_params = galprime.model_from_kde(self.kde, self.config, self.mag_kde, 
                                                           names=self.metadata["NAMES"])
        self.metadata.update(model_params)
        
        
    def get_background(self):
        
        # Convolve model with the PSF, and add the model to the background
        self.convolved_model = convolve2d(self.model, self.psf, mode='same')
        self.bgadded_cutout = self.convolved_model + self.background_cutout
        
        bg_stats = galprime.estimate_background_sigclip(self.bgadded_cutout, config=self.config)
        self.metadata["BG_MEAN"], self.metadata["BG_MED"], self.metadata["BG_STD"] = bg_stats
    
    def subtract_background(self):
        if self.config["DEBUG"]:
            print("Subtracting the background from the bgadded cutout")
        sm = galprime.SourceMask(self.convolved_model + self.background_cutout) 
        mask = sm.multiple(filter_fwhm=[1,3,5], tophat_size=[4,2,1])
        self.bgmask = mask
        
        bg_2D = galprime.background_2D(self.bgadded_cutout, self.bgmask, self.config["BOX_SIZE"], filter_size=self.config["FILTER_SIZE"])
        self.bg_model = bg_2D.background
        
    def mask_cutouts(self):
        if self.config["DEBUG"]:
            print("Masking all profiles")
        # Apply masking to both bgadded and bgsub, add these to the container, and update the metadata
        bgadded_masked = galprime.MaskedCutout(cutout=self.bgadded_cutout, config=self.config)
        bgadded_masked.mask_cutout()
        self.bgadded_masked = bgadded_masked.masked_cutout
        self.metadata.update(galprime.dict_extend(bgadded_masked.mask_info, extension="_BGA"))

        bgsub_masked = galprime.MaskedCutout(cutout=self.bgadded_cutout - self.bg_model, config=self.config)
        bgsub_masked.mask_cutout()
        self.metadata.update(galprime.dict_extend(bgsub_masked.mask_info, extension="_BGSUB"))            
        self.bgsub_masked = bgsub_masked.masked_cutout

            
    def extract_profiles(self):
        if self.config["DEBUG"]:
            print("Extracting all profiles")
        self.model_profile = galprime.isophote_fitting(self.convolved_model, config=self.config)
        self.bgadded_profile = galprime.isophote_fitting(self.bgadded_masked, config=self.config)
        self.bgsub_profile = galprime.isophote_fitting(self.bgsub_masked, config=self.config)
            
            
    def save_data(self):
        if self.config["DEBUG"]:
            print("Saving data")
    
    def plot_data(self, filename="out.png"):
        fig, ax = plt.subplots(2,5, facecolor="white")
        fig.set_figheight(5)
        fig.set_figwidth(12)
        
        scale = ZScaleInterval(contrast=0.5).get_limits(self.bgadded_cutout)
        
        ax[0][0].imshow(self.model, origin='lower', vmin=scale[0], vmax=scale[1], cmap="Greys")
        ax[0][1].imshow(self.convolved_model, origin='lower', vmin=scale[0], vmax=scale[1], cmap="Greys")
        ax[0][2].imshow(self.psf, origin='lower', cmap="Greys")
        ax[0][3].imshow(self.background_cutout, origin='lower', vmin=scale[0], vmax=scale[1], cmap="Greys")
        ax[0][4].imshow(self.convolved_model + self.background_cutout, origin='lower', vmin=scale[0], 
        vmax=scale[1], cmap="Greys")
        
        ax[1][0].imshow(self.bg_model, origin="lower", vmin=scale[0], vmax=scale[1], cmap="Greys")
        ax[1][1].imshow(self.bgmask, origin='lower', cmap="Greys")
        ax[1][2].imshow(self.bgadded_masked, origin='lower', vmin=scale[0], vmax=scale[1], cmap="Greys")
        ax[1][3].imshow(self.bgsub_masked, origin='lower', vmin=scale[0], vmax=scale[1], cmap="Greys")
        
        
        ax[0][0].set_title("Model")
        ax[0][1].set_title("Convolved Model")
        ax[0][2].set_title("PSF")
        ax[0][3].set_title("Background Cutout")
        ax[0][4].set_title("Combined Cutout")
        
        ax[1][0].set_title("Model Background")
        ax[1][1].set_title("Background Mask")
        ax[1][2].set_title("Masked BGAdded")
        ax[1][3].set_title("Masked BG-Subbed")
        ax[1][4].set_title("Profiles: Flux vs " + r'$R^{1/4}$')
        
        bare_prof = self.model_profile
        bgadded_prof = self.bgadded_profile
        bgsub_prof = self.bgsub_profile
        
        
        ax[1][4].plot(bare_prof.sma ** (1/4), np.log10(bare_prof.intens), color="red")
        ax[1][4].plot(bgadded_prof.sma ** (1/4), np.log10(bgadded_prof.intens), color="blue")
        ax[1][4].plot(bgsub_prof.sma ** (1/4), np.log10(bgsub_prof.intens), color="gold")

        for j in range(5):
            for i in range(2):
                ax[i][j].set_xticks([])
                ax[i][j].set_yticks([])
        
        ax[1][4].set_xticks(np.arange(0, np.max(bare_prof.sma ** (1/4))))
        
        plt.suptitle("ID: " + str(self.metadata["ID"]))
        plt.tight_layout()
        if filename is None:
            plt.show()
        else:
            plt.savefig(filename, dpi=150)
        
    
    def save_outputs(self, filename="GalPrimeContainer.fits", overwrite=True):
        print("Saving output")
        out_HDUList = fits.HDUList()
        head = fits.Header()
        
        for n in self.metadata:
            val = self.metadata[n]
            if type(val) == list:
                continue
            head[n] = self.metadata[n]
        
        out_HDUList.append(fits.PrimaryHDU(header=head))
        out_HDUList.append(fits.ImageHDU(data=self.convolved_model, name="MODEL"))
        out_HDUList.append(fits.ImageHDU(data=self.bgadded_cutout, name="BGA_IMG"))
        out_HDUList.append(fits.ImageHDU(data=self.bgsub_cutout, name="BGS_IMG"))
        out_HDUList.append(fits.ImageHDU(data=self.bgadded_masked, name="BGAMASKED"))
        out_HDUList.append(fits.ImageHDU(data=self.bgsub_masked, name="BGSMASKED"))
        out_HDUList.append(fits.ImageHDU(data=self.bg_model, name="BGMODEL"))
        out_HDUList.append(fits.BinTableHDU(data=galprime.table_from_isolist(self.model_profile), name="MOD_PROF"))
        out_HDUList.append(fits.BinTableHDU(data=galprime.table_from_isolist(self.bgadded_profile), name="BGA_PROF"))
        out_HDUList.append(fits.BinTableHDU(data=galprime.table_from_isolist(self.bgsub_profile), name="BGS_PROF"))
       
        print(len(out_HDUList), type(out_HDUList))
        
        out_HDUList.writeto(filename, overwrite=overwrite)
        

    def process_object(self, plot=False):
        t_init = time.time()
        self.generate_model()
        self.get_background()
        self.subtract_background()
        self.mask_cutouts()
        self.extract_profiles()
        self.save_data()
        
        self.metadata["T_ELPSD"] = time.time() - t_init

        if plot:
            outdir = self.config["OUT_DIR"] + "pngs/"
            if not os.path.isdir(outdir):
                os.mkdir(outdir)
            self.plot_data(filename=outdir + "obj_" + str(self.metadata["ID"]) + ".png")


import matplotlib as mpl

def load_pyplot_style():
    """ Load the style used for galprime plotting 
    Notably has a thicker border and tick marks 
    """
    mpl.rc('text', usetex=True)

    mpl.rcParams['xtick.major.size'] = 6
    mpl.rcParams['xtick.major.width'] = 2
    mpl.rcParams['xtick.minor.size'] = 4
    mpl.rcParams['xtick.minor.width'] = 1
    mpl.rcParams['ytick.major.size'] = 6
    mpl.rcParams['ytick.major.width'] = 3
    mpl.rcParams['ytick.minor.size'] = 4
    mpl.rcParams['ytick.minor.width'] = 1
    mpl.rcParams['axes.linewidth'] = 1.5

    mpl.rc('xtick', labelsize=18)
    mpl.rc('ytick', labelsize=18)

    mpl.rcParams['text.latex.preamble'] = [r'\boldmath']

    font = {'family' : 'serif',
        'weight': 'bold',
        'size': 20}

    mpl.rc('font', **font)






import galprime
import numpy as np
import os, time

from matplotlib import pyplot as plt

from astropy.io import fits
from astropy.convolution import Gaussian2DKernel

from numpy.random import randint
from scipy.signal import convolve2d

import galprime

from pebble import ProcessPool
from concurrent.futures import TimeoutError


class GalPrimeError(Exception):
    pass


class GPrime():
    """ Class to manage GalPrime simulations """
    
    def __init__(self, config=None, psfs=None, backgrounds=None, mag_kde=None, 
                 table_names=["ID","IMG", "BG_INDEX", "X", "Y", "RA", "DEC", "MAGS", "R50S", "NS", "ELLIPS", "MASS_MED",
                              "ZPHOT", "sfProb", "I_R50", "PA", "R50_PIX", "BG_MEAN", "BG_MED", "BG_STD",
                              "NSEG_BGA", "CENT_BGA", "N_MASKED_BGA", "P_MASKED_BGA", "NSEG_BGSUB", "CENT_BGSUB", 
                              "N_MASKED_BGSUB", "P_MASKED_BGSUB", "T_ELPSD"]):

        if config is None:
            self.config = galprime.default_config_params()
        elif type(config) == str:
            self.config = galprime.load_config_file(config)
        elif type(config) == dict:
            self.config = config
        else:
            raise GalPrimeError("Uncertain format for config file. File needs to be dict object or str filename.")
            sys.exit(1)
        
        self.binned_objects = None
        self.psfs = psfs
        self.backgrounds = backgrounds
        self.mag_kde = mag_kde

        self.table_names = table_names
        

    def load_bins(self):
        self.binned_objects = galprime.bin_catalog(self.config)
    
    
    def pipeline(self, max_bins=None, mag_kde=None, process_method=None,
                progress_bar=False, debug=False,):
        t_init = time.time()
        # Bin out catalog based on the catalog in the configuration file
        binned_objects = galprime.bin_catalog(self.config)
        max_bins = len(binned_objects.bins) if max_bins is None else max_bins
        verbose = self.config["VERBOSE"]

        if verbose:
            print("Running GalPRIME, version", galprime.__version__)
            print("-" + str(len(binned_objects.bins)), "bins in total. Running GalPRIME on", max_bins, "bins total")
            if self.backgrounds is not None:
                print("-Backgrounds were supplied:", len(self.backgrounds.cutouts), "in total.")


        # Generate the full filestructure, including containers for the individual profiles, and the medians
        # The directories object contains all the required directories in a neat dict format
        directories = galprime.gen_filestructure(self.config["OUT_DIR"])

        # Run through bins and process using the method
        for i in range(max_bins):
            current_bin = binned_objects.bins[i]
            columns = current_bin.columns()
            index_prefix = galprime.gen_index_prefix(current_bin.bin_params, self.config)

            
            # Generate the object KDE that will be used by this bin, and save it to a pickle file 
            kde = galprime.object_kde(columns)
            galprime.save_pickle(kde, directories["ADDITIONAL"] + index_prefix + "_kde.dat")


            # Set up the job pool
            job_list, job_results = [], []
            if verbose:
                print("Working through", self.config["N_MODELS"], "models for bin:", current_bin.bin_params, 
                      "with", self.config["CORES"], "cores." )
            
            # with ProcessPool(max_workers=self.config["CORES"]) as pool:
            #     for i in range(self.config["N_MODELS"]):
            #         job_list.append(pool.schedule(process_method,
            #                                     args=(self, current_bin, kde),
            #                                     timeout=self.config["TIME_LIMIT"]))
            # # Now get all the results
            # for i, job in enumerate(job_list):
            #     try:
            #         job_results.append(job.result())
            #     except Exception as error:
            #         print(" ", error.args, i)

            def task_done(future):
                try:
                    result = future.result()  # blocks until results are ready
                    job_results.append(result)
                except TimeoutError as error:
                    print("Function took longer than %d seconds" % error.args[1])
                except Exception as error:
                    print("Function raised %s" % error)
                    if debug:
                        print(error.traceback)  # traceback of the function


            with ProcessPool(max_workers=self.config["CORES"]) as pool:
                for index in range(self.config["N_MODELS"]):
                    future = pool.schedule(process_method, args=(self, current_bin, kde),
                                           timeout=self.config["TIME_LIMIT"])
                    future.add_done_callback(task_done)


            # Clean the jobs of any that just so happened to fail and return junk profiles
            cleaned_containers = []
            for container in job_results:
                good = True
                for prof in (container.model_profile, container.bgadded_profile, container.bgsub_profile):
                    if len(prof) == 0:
                        good = False
                        break
                if good:
                    cleaned_containers.append(container)
            print(" . Profiles cleaned", len(job_results), len(cleaned_containers))
            job_results = cleaned_containers
            

            # Now save all of the data into nice outputs
            bare_hdulist, bgadded_hdulist, bgsub_hdulist = fits.HDUList(), fits.HDUList(), fits.HDUList()
     
            for i, container in enumerate(job_results):
                bare_hdulist.append(fits.BinTableHDU(data=galprime.table_from_isolist(container.model_profile)))
                bgadded_hdulist.append(fits.BinTableHDU(data=galprime.table_from_isolist(container.bgadded_profile)))
                bgsub_hdulist.append(fits.BinTableHDU(data=galprime.table_from_isolist(container.bgsub_profile)))

            try:
                bare_hdulist.writeto(directories["BARE_PROFILES"] + index_prefix + "bare.fits", overwrite=True)
                bgadded_hdulist.writeto(directories["BGADDED_PROFILES"] + index_prefix + "bgadded.fits", overwrite=True)
                bgsub_hdulist.writeto(directories["BGSUB_PROFILES"] + index_prefix + "bgsub.fits", overwrite=True)
            except Exception as error:
                print(" . Error Generating Data Products (Not good!):", error.args)

            try:
                bare_profiles = [galprime.table_from_isolist(container.model_profile) for container in job_results]
                bare_median_table = galprime.boostrap_median(bare_profiles)
                bare_median_table.write(directories["BARE_MEDIANS"] + index_prefix + "medians.fits", overwrite=True)
            except Exception as error:
                print(" . Error Bootstrapping Bare:", error.args)
            try:    
                bgadded_profiles = [galprime.table_from_isolist(container.bgadded_profile) for container in job_results]
                bgadded_median_table = galprime.boostrap_median(bgadded_profiles)
                bgadded_median_table.write(directories["BGADDED_MEDIANS"] + index_prefix + "medians.fits", overwrite=True)
            except Exception as error:
                print(" . Error Bootstrapping Bgadded:", error.args)
            try:
                bgsub_profiles = [galprime.table_from_isolist(container.bgsub_profile) for container in job_results]
                bgsub_median_table = galprime.boostrap_median(bgsub_profiles)
                bgsub_median_table.write(directories["BGSUB_MEDIANS"] + index_prefix + "medians.fits", overwrite=True)
            except Exception as error:
                print(" . Error Bootstrapping Bgsub:", error.args)

            try:
                output_datatable = galprime.construct_infotable(job_results, self.table_names)
                output_datatable.write(directories["ADDITIONAL"] + index_prefix + "_data.fits", 
                                    format="fits", overwrite=True)
            except Exception as error:
                print(" . Error with saving the output table (compare metadata with table_names):", error.args)

            if verbose:
                print("Finished. Time elapsed:", np.round((time.time() - t_init) / 60, 3), "minutes.")

            # ONLY in debugging, you can return the GalPrimeContainer list for the first bin.
            if debug:
                return job_list

def gprime_single(gprime_obj, current_bin, kde):
    metadata = {}
    metadata["NAMES"] = current_bin.object_column_names
    metadata["BIN_PARAMS"] = current_bin.bin_params

    # Select a background and add the background info to the metadata
    bg_index = randint(len(gprime_obj.backgrounds.cutouts))
    metadata["BG_INDEX"] = bg_index
    metadata.update(gprime_obj.backgrounds.cutout_data[bg_index])

    # Get the best PSF is the user has supplied PSFs to the pipeline
    # Otherwise, just use a simple Gaussian2D kernel from astropy
    if gprime_obj.psfs is None:
        psf = Gaussian2DKernel(x_stddev=1.5, y_stddev=1.5)
    else:
        psf = galprime.get_closest_psf(gprime_obj.psfs, metadata["RA"], metadata["DEC"])

    # Generate our container
    container = galprime.GalPrimeContainer(config=gprime_obj.config, kde=kde, psf=np.copy(psf),
                                    background_cutout=np.copy(gprime_obj.backgrounds.cutouts[bg_index]), 
                                    metadata=metadata)
    container.process_object(plot=False)
    return container



from matplotlib import pyplot as plt
from astropy.table import Table
from numpy import copy, log10
import galprime


def adjust_profile(x, y, m_0=27, arcconv=0.168):

    x_new, y_new = copy(x), copy(y)
    y_new = -2.5 * log10(y / (arcconv ** 2)) + m_0

    return x_new, y_new


# def single_bin_plot(table_arrays, colours=None, ind_profile_alpha=1, medians=False, to_sb=True,
#                     bootstraps=False, ylim=(30, 20),
#                     xlabel=None, ylabel=None, output="", dpi=150):

#     plt.figure(figsize=(8, 6))

#     for i in range(0, len(table_arrays)):
#         for j in range(0, len(table_arrays[i])):
#             this_prof = table_arrays[i][j]

#             sma, intens = this_prof["sma"], this_prof["intens"]
#             sma, intens = adjust_profile(sma, intens)

#             plt.plot(sma, intens, c=colours[i], alpha=ind_profile_alpha, lw=1)

#         if medians:
#             interps = tbridge.as_interpolations(table_arrays[i])
#             median_sma, median_interp = tbridge.get_median(interps, tbridge.bin_max(table_arrays[i]))
#             median_intens = median_interp(median_sma)

#             if to_sb:
#                 median_sma, median_intens = adjust_profile(median_sma, median_intens)

#             plt.plot(median_sma, median_intens, lw=3, c=colours[i])

#     if ylim is None:
#         pass
#     else:
#         plt.ylim(ylim[0], ylim[1])
#     plt.xlim(1,)
#     plt.xscale("log")

#     if xlabel is not None:
#         plt.xlabel(xlabel)
#     if ylabel is not None:
#         plt.ylabel(ylabel)

#     plt.tight_layout()

#     if output == "":
#         plt.show()
#     else:
#         plt.savefig(output, dpi=dpi)


def single_prof(x, y, error=None):
    plt.figure(figsize=(8, 6))

    if error is None:
        plt.plot(x, y, color="orange", lw=2)
    else:
        plt.errorbar(x, y, yerr=error)

    plt.tight_layout()
    plt.show()



import os
import numpy as np
from scipy.interpolate import interp1d

from matplotlib import pyplot as plt

from astropy.table import Table


def max_sma(table_list, sma_key="sma"):
    """Get the maximum semimajor axis for a table list

    :param table_list: _description_
    :type table_list: _type_
    :param sma_key: _description_, defaults to "sma"
    :type sma_key: str, optional
    :return: _description_
    :rtype: _type_
    """
    global_max = 0
    for tab in table_list:
        this_max = np.max(tab[sma_key])
        if this_max > global_max:
            global_max = this_max
    return global_max



def interp_median(xs, interp_list):
    """Get the median for a list of interp1D object profiles

    :param xs: The x slices to get the median at
    :type xs: array_like
    :param interp_list: The list of interp1D objects (individual profiles)
    :type interp_list: list(scipy.interp1D)
    :return: A 1D interpolation of the median profile.
    :rtype: scipy.interpolate.interp1D
    """
    median_prof = []
    for x in xs:
        median_prof.append(np.median([interp(x) for interp in interp_list]))
    return interp1d(xs, median_prof, bounds_error=False, fill_value="extrapolate")


def array2D_median(array2D):
    """ Get the median of a 2D array of profiles (used in bootstrap_median()) """
    return np.nanmedian(array2D, axis=0)


def boostrap_median(table_list, iterations=500, x_step=0.5, plot_dir=None, plot_prefix=""):
    """ Generate an Astropy Table object containing the values from bootstrapping of a set of profiles.

    :param table_list: A set of tables (outputs from galprime.isophote_fitting, processed using the to_table() 
        method). 
    :type table_list: list(Table())
    :param iterations: The number of bootstrap iterations to run, defaults to 500
    :type iterations: int, optional
    :param x_step: The step between samples along the x-axis, defaults to 0.5
    :type x_step: float, optional
    :param plot_dir: Directory to generate output plots where if None, no plots are made, defaults to None
    :type plot_dir: str, optional
    :param plot_prefix: Prefix to include in the output plots. This is helpful if you have a 
                        bin and want to include which bin the plots are coming from, defaults to ""
    :type plot_prefix: str, optional
    :return: The table of 
    :rtype: astropy Table object
    """
    
    # We will start by making a 2D image with a shared x-axis, so we can sample WAY FASTER than using interp1D objects
    interp_list = [interp1d(tab["sma"], tab["intens"], bounds_error=False, fill_value="extrapolate") for tab in table_list]
    
    x_slices = np.arange(0, max_sma(table_list), x_step)
    full_array = []
    for n in interp_list:
        full_array.append(n(x_slices))
    full_array = np.asarray(full_array)     # This is now a 2D array (basically an image)

    median_prof = array2D_median(full_array)
    bs_1sig_u, bs_1sig_l = [], []
    bs_2sig_u, bs_2sig_l = [], []
    bs_3sig_u, bs_3sig_l = [], []

    lower_index_1sig, upper_index_1sig = int(iterations * 0.159), int(iterations * 0.841)
    lower_index_2sig, upper_index_2sig = int(iterations * 0.023), int(iterations * 0.977)
    lower_index_3sig, upper_index_3sig = int(iterations * 0.002), int(iterations * 0.998)


    boostrapped_medians = []
    indices = np.arange(0, len(table_list))

    # Bootsrap our population, generate a new 2D array, and add the median of that to our list
    for iter in range(iterations):
        pop_indices = np.random.choice(indices, size=len(table_list), replace=True)
        pop_array2D = full_array[pop_indices]
        boostrapped_medians.append(array2D_median(pop_array2D))
    
    boostrapped_medians = np.asarray(boostrapped_medians)
    
    # Now go through the y slices, and get the bootstrap values at each sigma interval
    for i in range(boostrapped_medians.shape[1]):
        y_slice = np.sort(boostrapped_medians[:, i])
        bs_1sig_l.append(y_slice[lower_index_1sig])
        bs_1sig_u.append(y_slice[upper_index_1sig])
        bs_2sig_l.append(y_slice[lower_index_2sig])
        bs_2sig_u.append(y_slice[upper_index_2sig])
        bs_3sig_l.append(y_slice[lower_index_3sig])
        bs_3sig_u.append(y_slice[upper_index_3sig])

    colnames = ["SMA", "MEDIAN", "L_1SIG", "U_1SIG", "L_2SIG", "U_2SIG", "L_3SIG", "U_3SIG",]
    bootstrap_table = Table([x_slices, median_prof, bs_1sig_l, bs_1sig_u, bs_2sig_l, bs_2sig_u, bs_3sig_l, bs_3sig_u],
                            names=colnames)
    
    # Here are some nice plotting routines 
    if plot_dir is not None:
        if not os.path.isdir(plot_dir):
            os.makedirs(plot_dir)
        

        plt.figure(facecolor="white")
        plt.imshow(np.log10(full_array), vmin=-6, vmax=1)
        plt.yticks([])
        plt.xticks([])
        plt.title("Input Profiles")
        plt.tight_layout()
        plt.savefig(plot_dir + plot_prefix + "_inputs.png", dpi=150)


        plt.figure(facecolor="white")
        plt.imshow(np.log10(boostrapped_medians), vmin=-6, vmax=1)
        plt.yticks([])
        plt.xticks([])
        plt.title("Bootstrapped Median Profiles")
        plt.tight_layout()
        plt.savefig(plot_dir + plot_prefix + "_bootstrapped.png", dpi=150)


        plt.figure(facecolor="white")
        for i in range(full_array.shape[0]):
            plt.plot(x_slices, full_array[i], color="grey")
        
        plt.plot(x_slices, median_prof, color="red")

        plt.yscale("log")
        plt.ylim(1e-6, 10)
        plt.xlabel(r'$SMA$ [pix]')
        plt.ylabel("I(R)")

        plt.title("Profiles")
        plt.tight_layout()
        plt.savefig(plot_dir + plot_prefix + "_profiles.png", dpi=150)


        plt.figure(facecolor="white")
        plt.plot(x_slices, median_prof, color="Red")
        plt.plot(x_slices, bs_3sig_u, color="orange")
        plt.plot(x_slices, bs_3sig_l, color="orange")
        plt.plot(x_slices, bs_2sig_u, color="black")
        plt.plot(x_slices, bs_2sig_l, color="black")
        plt.plot(x_slices, bs_1sig_u, color="grey")
        plt.plot(x_slices, bs_1sig_l, color="grey")

        plt.yscale("log")
        plt.ylim(1e-6, 10)
        plt.xlabel(r'$SMA$ [pix]')
        plt.ylabel("I(R)")

        plt.title("Profiles")
        plt.tight_layout()
        plt.savefig(plot_dir + plot_prefix + "_boostrapped_medians.png", dpi=150)

    return bootstrap_table
